# ......................................................................................
# MIT License

# Copyright (c) 2024-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................


import pandas as pd


from .fileobject import FileObject

class CSVFile(FileObject):
  # ----------------------------------------------------------------------------------
  def __init__(self, filename, parent_folder=None, error_template=None):
    super(CSVFile, self).__init__(filename, parent_folder, error_template, "txt")
  # ----------------------------------------------------------------------------------
  def load(self, filename=None, delimiter=",", encoding="utf-8"):
    sFileName = self._useFileName(filename)
    dResult = pd.read_csv(sFileName, delimiter=delimiter, encoding=encoding)
    return dResult
  # ----------------------------------------------------------------------------------
  def save(self, obj, filename=None):
    # TODO Save numpy array to pandas dataframe then to CSV
    pass

  # ----------------------------------------------------------------------------------


# ......................................................................................
# MIT License

# Copyright (c) 2024-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................

import os
import json
import glob
from .fileobject import FileObject

class JSONFile(FileObject):
  # ----------------------------------------------------------------------------------
  def __init__(self, filename, parent_folder=None, error_template=None):
    super(JSONFile, self).__init__(filename, parent_folder, error_template, "json")
  # ----------------------------------------------------------------------------------
  def load(self, filename=None):
    filename = self._useFileName(filename)

    dResult = None
    if os.path.exists(filename):
      with open(filename) as oFile:
        sJSON = oFile.read()
      dResult = json.loads(sJSON)
    else:
      if self.error_template is not None:
        raise Exception(self.error_template % filename)

    return dResult
  # ----------------------------------------------------------------------------------
  def save(self, obj, filename=None, is_sorted_keys=True):
    filename = self._useFileName(filename)

    if obj is not None:
      sJSON = json.dumps(obj, default=lambda o: o.__dict__, sort_keys=is_sorted_keys, indent=4)
      with open(filename, "w") as oFile:
        oFile.write(sJSON)
        oFile.close()
  # ----------------------------------------------------------------------------------
  @property
  def files(self, is_full_path=True):
    oResult = []
    if (self.parent_folder is not None):
      oJSONFiles = glob.glob(os.path.join(self.parent_folder, '*.json'))
      oJSONFiles = sorted(oJSONFiles, key=os.path.getmtime)

      for sJSONFile in oJSONFiles:
        if is_full_path:
          oResult.append(os.path.join(self.parent_folder, sJSONFile))
        else:
          oResult.append(sJSONFile)
    return oJSONFiles
  # ----------------------------------------------------------------------------------


from radnn import FileSystem
from radnn.experiment import MLExperimentConfig


oSys = FileSystem("MLConfig", "MLModels", "MLData")

oConfig = MLExperimentConfig(expt_number=3).assign(
        {
            "ModelName": "QPEDS_MLP"
           ,"DNN.InputFeatures": 72
           ,"DNN.LayerNeurons": [288, 4]
           ,"DNN.Classes": 4
           ,"Training.MaxEpoch": 200
           ,"Training.BatchSize": 160
           ,"Training.LearningRate": 0.2
           ,"Experiment.RandomSeed": 2022
         }
).save_config(oSys)
print(oConfig)

print("."*40, "Loading", "."*40)
oConfig = MLExperimentConfig().load_config(oSys, "QPEDS_MLP_03")

print(oConfig)


from radnn import FileSystem

TEST_NUM = 1

if TEST_NUM == 1:
  oSys = FileSystem("MLConfig", "MLModels", "MLData", "DNN")
  oSys.save_setup()
elif TEST_NUM == 2:
  oSys = FileSystem()
  print(oSys)
elif TEST_NUM == 3:
  oSys = FileSystem(setup_filename="test.fsys")
  print(oSys)





# ......................................................................................
# MIT License

# Copyright (c) 2018-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ......................................................................................

import os
from radnn.core import system_name
from radnn.system.filestore import FileStore
from radnn.system.files import JSONFile




# =======================================================================================================================
class FileSystem(object):
  # --------------------------------------------------------------------------------------------------------
  def __init__(self, config_folder=None, model_folder=None, dataset_folder=None, model_group=None, must_exist=False, setup_filename=None):
    '''
    Initializes the file system settings for an experiment
    :param config_folder: The folder that contains the experiment hyperparameter files.
    :param model_folder: The main folder where the models are stored.
    :param dataset_folder: The main folder under which dataset folder are stores.
    :param model_group: A group of models under the main folder will be used instead.
    :param setup_filename: The filename of the setup file.
    :param must_exist: False: Auto-creates the directories on the file system | True: Raises an error if it does not exist.
    '''

    self.setup_filename = setup_filename
    if self.setup_filename is None:
      self.setup_filename = system_name() + ".fsys"

    if os.path.exists(self.setup_filename):
      oFile = JSONFile(self.setup_filename)
      dSetup = oFile.load()
      config_folder = dSetup["Configs"]
      model_folder = dSetup["Models"]
      dataset_folder = dSetup["Datasets"]
      if model_group is None:
        if "ModelGroup" in dSetup:
          model_group = dSetup["ModelGroup"]
    else:
      dSetup = dict()
      dSetup["Configs"] = config_folder
      dSetup["Models"] = model_folder
      dSetup["Datasets"] = dataset_folder

    if model_group is not None:
      config_folder   = os.path.join(config_folder, model_group)
      model_folder    = os.path.join(model_folder, model_group)

    # ...................... | Fields | ......................
    self.setup = dSetup
    self.configs   = FileStore(config_folder, must_exist=must_exist)
    self.models    = FileStore(model_folder, must_exist=must_exist)
    self.datasets  = FileStore(dataset_folder, must_exist=must_exist)
    # ........................................................
  # --------------------------------------------------------------------------------------------------------
  def save_setup(self):
    oFile = JSONFile(self.setup_filename)
    oFile.save(self.setup)
  # --------------------------------------------------------------------------------------------------------
  def __str__(self)->str:
    sResult = f"  ConfigsFolder: \"{self.configs.base_folder}\",\n"
    sResult += f"  ModelsFolder: \"{self.models.base_folder}\",\n"
    sResult += f"  DatasetsFolder: \"{self.datasets.base_folder}\"\n"
    sResult = "{\n" + sResult + "}"
    return sResult
  # --------------------------------------------------------------------------------------------------------
  def __repr__(self)->str:
    return self.__str__()
  # --------------------------------------------------------------------------------------------------------
# =======================================================================================================================



# ......................................................................................
# MIT License

# Copyright (c) 2023-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................

import os
import json


# --------------------------------------------------------------------------------------
def experiment_code(config_dict):
  if ("Experiment.BaseName" in config_dict) and ("Experiment.Number" in config_dict):
    sBaseName = config_dict["Experiment.BaseName"]
    nNumber = int(config_dict["Experiment.Number"])
    sVariation = None
    if "Experiment.Variation" in config_dict:
      sVariation = config_dict["Experiment.Variation"]
    nFoldNumber = None
    if "Experiment.FoldNumber" in config_dict:
      nFoldNumber = config_dict["Experiment.FoldNumber"]

    sCode = "%s_%02d" % (sBaseName, nNumber)
    if sVariation is not None:
      sCode += ".%s" % str(sVariation)
    if nFoldNumber is not None:
      sCode += "-%02d" % int(nFoldNumber)
  else:
    raise Exception("Invalid experiment configuration. Needs at least two keys 'Experiment.BaseName'\n"
                  + "and `Experiment.Number`.")

  return sCode
# --------------------------------------------------------------------------------------





# =========================================================================================================================
class MLExperimentConfig(dict):
  # --------------------------------------------------------------------------------------
  def __init__(self, filename=None, expt_base_name=None, expt_number=None, expt_variation=None, expt_fold=None, hyperparams=None):
    self["Experiment.BaseName"] = expt_base_name
    self.filename = filename
    if filename is not None:
      self.load()

    if expt_number is not None:
      self["Experiment.Number"] = expt_number
    if expt_variation is not None:
      self["Experiment.Variation"] = expt_variation
    if expt_fold is not None:
      self["Experiment.FoldNumber"] = expt_fold

    if hyperparams is not None:
      self.assign(hyperparams)
  # --------------------------------------------------------------------------------------
  @property
  def experiment_code(self):
    return experiment_code(self)
  # --------------------------------------------------------------------------------------
  def load(self, filename, must_exist=False):
    if filename is None:
      filename = self.filename

    # reading the data from the file
    if os.path.exists(filename):
      with open(filename) as oFile:
        sConfig = oFile.read()
        self.setDefaults()
        dConfigDict = json.loads(sConfig)

      for sKey in dConfigDict.keys():
        self[sKey] = dConfigDict[sKey]
    else:
      if must_exist:
        raise Exception("Experiment configuration file %s is not found." % filename)
    return self
  # --------------------------------------------------------------------------------------
  def assign(self, config_dict):
    for sKey in config_dict.keys():
      self[sKey] = config_dict[sKey]

    if (self["Experiment.BaseName"] is None) and ("ModelName" in config_dict):
      self["Experiment.BaseName"] = config_dict["ModelName"]
    if ("DatasetName" in config_dict):
      self["Experiment.BaseName"] += "_" + config_dict["DatasetName"]
    return self
  # --------------------------------------------------------------------------------------
  def save(self, filename):
    if filename is not None:
      self.filename = filename

    sJSON = json.dumps(self, default=lambda o: o.__dict__, sort_keys=False, indent=4)
    with open(self.filename, "w") as oFile:
      oFile.write(sJSON)
      oFile.close()

    return self

  # --------------------------------------------------------------------------------------
  def save_config(self, filesystem, filename_only=None):
    if filename_only is None:
      filename_only = experiment_code(self)

    sFileName = filesystem.configs.file(filename_only + ".json")
    return self.save(sFileName)

  # --------------------------------------------------------------------------------------
  def load_config(self, filesystem, filename_only):
    sFileName = filesystem.configs.file(filename_only + ".json")
    return self.load(sFileName)
  # --------------------------------------------------------------------------------------
  def setDefaults(self):
    pass
  # --------------------------------------------------------------------------------------
  def __str__(self)->str:
    sResult = ""
    for sKey in self.keys():
      sResult += f'  {sKey}: \"{self[sKey]}\",\n'

    sResult = "{\n" + sResult + "}"
    return sResult
  # --------------------------------------------------------------------------------------------------------
  def __repr__(self)->str:
    return self.__str__()
  # --------------------------------------------------------------------------------------------------------

# =========================================================================================================================        


  


# ......................................................................................
# MIT License

# Copyright (c) 2018-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................


import os
import shutil
import glob
import sys
if (sys.version_info.major == 3) and (sys.version_info.minor <= 7):
  import pickle5 as pickle
else:
  import pickle

from radnn.system.files import JSONFile
from radnn.system.files import PickleFile
from radnn.system.files import TextFile
from radnn.system.files import CSVFile

from radnn.core import is_opencv_installed
if (is_opencv_installed()):
  from radnn.system.files.imgfile import PNGFile




# =======================================================================================================================
class FileStore(object):
  # --------------------------------------------------------------------------------------------------------
  def __init__(self, base_folder, is_verbose=False, must_exist=False):
    #.......................... |  Instance Attributes | ............................
    self.base_folder = base_folder
    if not os.path.exists(self.base_folder):
      if must_exist:
        raise Exception(f"File store folder {self.base_folder} does not exist.")
      else:
        os.makedirs(self.base_folder)

    self.is_verbose  = is_verbose
    self.json = JSONFile(None, parent_folder=self.base_folder)
    self.obj = PickleFile(None, parent_folder=self.base_folder)
    self.text = TextFile(None, parent_folder=self.base_folder)
    self.csv = CSVFile(None, parent_folder=self.base_folder)
    if (is_opencv_installed()):
      self.img = PNGFile(None, parent_folder=base_folder)
    self.donefs = None
    #................................................................................
  # --------------------------------------------------------------------------------------------------------
  @property
  def has_files(self):
    return not self.is_empty
  # --------------------------------------------------------------------------------------------------------
  @property
  def is_empty(self):
    bExists = not os.path.exists(self.base_folder)
    if not bExists:
      bExists = len(os.listdir(self.base_folder)) > 0
    return not bExists
  # --------------------------------------------------------------------------------------------------------
  def exists_folder(self, filename):
    sFullPath = os.path.join(self.base_folder, filename)
    return os.path.exists(sFullPath)
  # --------------------------------------------------------------------------------------------------------
  def exists(self, filename):
    sFullFilePath = os.path.join(self.base_folder, filename)
    return os.path.isfile(sFullFilePath)
  # --------------------------------------------------------------------------------------------------------
  def subfs(self, subfolder_name, must_exist=False):
    return FileStore(self.subpath(subfolder_name), must_exist=must_exist)
  # --------------------------------------------------------------------------------------------------------
  def subpath(self, subfolder_name):
    if os.path.sep == "\\":
      if subfolder_name.find(os.path.sep) < 0:
        subfolder_name = subfolder_name.replace("/", "\\")
    return self.folder(subfolder_name)
  # --------------------------------------------------------------------------------------------------------
  def folder(self, folder_name):
    sFolder = os.path.join(self.base_folder, folder_name)
    if not os.path.exists(sFolder):
      os.makedirs(sFolder)

    return sFolder
  # --------------------------------------------------------------------------------------------------------
  def file(self, file_name, file_ext=None):
    if file_ext is not None:
      if file_ext.find(".") < 0:
        file_name += "." + file_ext
      else:
        file_name += file_ext
    return os.path.join(self.base_folder, file_name)
  # --------------------------------------------------------------------------------------------------------
  def entries(self):
    return os.listdir(self.base_folder)
  # --------------------------------------------------------------------------------------------------------
  def list_files(self, file_matching_pattern, is_full_path=True, is_removing_extension=False, sort_filename_key=None):
    sEntries = glob.glob1(self.base_folder, file_matching_pattern)
    if is_removing_extension:
      oFileNamesOnly = []
      for sEntry in sEntries:
        sFileNameOnly, _ = os.path.splitext(sEntry)
        oFileNamesOnly.append(sFileNameOnly)
      sEntries = sorted(oFileNamesOnly, key=sort_filename_key)

    if is_full_path:
      oResult = [os.path.join(self.base_folder, x) for x in sEntries]
    else:
      oResult = [x for x in sEntries]

    return oResult
  # --------------------------------------------------------------------------------------------------------
  @property
  def lis_folders(self, is_full_path=True):
    sResult = []
    for sFolder in os.listdir(self.base_folder):
      sFullPath = os.path.join(self.base_folder, sFolder)
      if os.path.isdir(sFullPath):
        if is_full_path:
          sResult.append(sFullPath)
        else:
          sResult.append(sFolder)
    return sResult
  # --------------------------------------------------------------------------------------------------------
  def dequeue_file(self, file_matching_pattern, is_full_path=True, archive_folder_name=".done"):
    if self.donefs is None:
      self.donefs = self.subfs(archive_folder_name)
    oQueue = self.files(file_matching_pattern, is_full_path=False)

    if len(oQueue) == 0:
      sFileName = None
    else:
      sFileName = oQueue[0]
      shutil.move(os.path.join(self.base_folder, sFileName), os.path.join(self.donefs.base_folder, sFileName))

    if is_full_path and (sFileName is not None):
      sFileName = os.path.join(self.donefs.base_folder, sFileName)
    return sFileName
  # --------------------------------------------------------------------------------------------------------
  def purge_done(self):
    #//TODO: Remove from the current filestore all files that are moved into the .done filestore
    pass
  # ----------------------------------------------------------------------------------
  def __repr__(self)->str:
    return self.base_folder
  # --------------------------------------------------------------------------------------------------------
  def __str__(self)->str:
    return self.base_folder
  # --------------------------------------------------------------------------------------------------------
# ======================================================================================================================


# .................... Tests ....................
if __name__ == "__main__":
  VISUALIZE = False
  import numpy as np

  # Lazy initialize a file store folder
  oFS = FileStore("C:\MLData\SomeDataSet")

  # Serialize data into a sub folder
  oData = np.asarray([1, 2, 3]).astype(np.float32)
  print(oFS.subfs("test"))
  oFS.subfs("test").obj.save(oData, "sample1.pkl")

  # Get paths
  print("==== Paths ====")
  print(oFS.subfs("test").file("sample1.png"))
  print(oFS.folder("test"))

  # Deserialize data from a sub folder
  oLoadedData = oFS.subfs("test").obj.load("sample1.pkl")
  print(oLoadedData)

  # List files
  print("==== Files ====")
  sFiles = oFS.subfs("test").list_files("*.pkl", True)
  print("\n".join(sFiles))

  # Load a PNG file
  if (is_opencv_installed()):
    oPng = FileStore("C:\\UX").subfs("PNG").img.load("Ideas.png")
    print(oPng.shape)

    if VISUALIZE:
      import matplotlib.pyplot as plt
      plt.imshow(oPng)
      plt.show()



